/* Grammar for PDDL */
/* pddlgram.y	-- Hans-Peter Stoerr Thu Jan  7 1999
 * $Id: pddlgram.y,v 4.4 2003/12/04 07:23:37 stoerr Exp $
 */

%{
#define YYERROR_VERBOSE 1
#define YYDEBUG 1

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <string>
#include <functional>

using namespace std;  
  
#include "general.h"
#include "pddldomain.h"
#include "misc/mystl.h"  

int yylex();
extern int      yymylineno;
extern FILE*    yyin;
static void yyerror(char *);
void printerrorlocation();

PddlDatabase*   yydatabase;

static PddlDomain*	domain=NULL; // parsed domain
static string		addendumname=""; // parsed addendum
static PddlProblem*	problem=NULL; // parsed problem
static PddlSituation*	situation=NULL; // parsed situation
static PddlAction*	action=NULL; // parsed action
static PddlAxiom*	axiom=NULL; // parsed axiom
;

static StringStringPairList* PddlVarListAddType(StringList* varlist, string type);
static StringStringPairList* PddlVarListAddType(StringList* varlist, string* type, StringStringPairList* rest);

static StringStringMap* PddlVarSetAddType(StringList* varlist, string type);
static StringStringMap* PddlVarSetAddType(StringList* varlist, string* type, StringStringMap* rest);

static string*  PddlEitherID(StringList* typelist);

static Atomtype PddlMakeAtomtype(StringStringPairList* typedvarlist);

%}

%union
{
  string *anstring;
  StringSet *stringset;
  StringList *stringlist;
  StringStringMap *stringstringmap;
  StringStringListMap *stringstringlistmap;
  StringStringPairList *stringstringpairlist;
  TermList *termlist;
  TermSet *termset;
  Term *term;
  StringAtomtypeMap *stringatomtypemap;
  PddlAxiom *axiom;
  PddlAction *action;
}

%token ACTION AXIOM CONSTANTS CONTEXT DDOMAIN EITHER EXTENDS EFFECT EFFECTPOS
%token EFFECTNEG FLUENT GOAL INIT IMPLIES LENGTH CONSTANTS PARALLEL
%token PARAMETERS PRECONDITION PREDICATES REQUIREMENTS SAFETY SERIAL
%token DSITUATION TIMELESS TYPES VARS ADDENDUM AND DEFINE DOMAINSTR EXISTS
%token FORALL IMPLY NOT OR PROBLEM SITUATION WHEN ISGOAL ISINITIAL CONSTRAINT LIMITED


%token <anstring> NAME VARNAME REQUIREMENT
%type <anstring> situation type
%type <stringlist> namelist varlist varornamelist
%type <stringset> reqsset probrequiredef
%type <stringstringmap> typedvarset typednameset evtlvars objectdecl
%type <stringstringpairlist> typedvarlist
%type <stringatomtypemap> atomtypemap
%type <termset> literalset goaldescrset effectset init
%type <term> atom literal effect goaldescr goal equality
%type <axiom> axiomdef
%type <action> actiondef

%start Pddlfile

%%

Pddlfile:       definition Pddlfile
|		/* empty */
;

definition:	domaindef
|		addendumdef
|		problemdef
|		situationdef
;

/* ================================================================ */
  
domaindef:	'(' DEFINE '(' DOMAINSTR NAME ')'
				{ domain=new PddlDomain($5); }
		domaindecls ')' { yydatabase->RegisterDomain(domain); domain=NULL; }
;

domaindecls:	domaindecl domaindecls | /*empty*/ ;

domaindecl:	extensiondef
|		requiredef
|		typesdef
|		constantsdef
|		predicatesdef
|		timelessdef
|		safetydef
|		constraintdef
|		structuredef
;		
		
extensiondef:	'(' EXTENDS NAME ')' { domain->Extends($3); };

requiredef:	'(' REQUIREMENTS reqsset ')' { domain->Requires($3); };

typesdef:	'(' TYPES typednameset ')' { domain->SetSubtypes($3); };

constantsdef:	'(' CONSTANTS typednameset ')' { domain->SetConstants($3); };

predicatesdef:	'(' PREDICATES atomtypemap ')' { domain->SetAtomtypes($3); };

timelessdef:	'(' TIMELESS literalset ')'	{ domain->SetTimeless($3); };

safetydef:	'(' SAFETY goaldescr ')'	{ domain->AddSafety($3); };

constraintdef:	'(' CONSTRAINT goaldescr ')'	{ domain->AddConstraint($3); };

structuredef:	actiondef			{ domain->AddAction($1); }
|		axiomdef			{ domain->AddAxiom($1); }
;

actiondef:	'(' ACTION NAME			{ action=new PddlAction($3); }
                evtlactionparms
		actiondefbody ')'		{ $$=action; action=NULL; }
;

evtlactionparms: PARAMETERS '(' typedvarlist ')'	{ action->SetParameters($3); }
|               /* empty */
;

/* :TODO: the fluent calculus support lacks multiple SUA per action. */
actiondefbody:	VARS '(' typedvarset ')' actiondefbody	{ action->SetVars($3); }
|		PRECONDITION goaldescr actiondefbody	{ action->SetPrecondition($2); }
|		EFFECT effect actiondefbody		{ action->SetEffect($2); }
|		EFFECTPOS literalset actiondefbody		{ action->SetPosEffect($2); }
|		EFFECTNEG literalset actiondefbody		{ action->SetNegEffect($2); }
|		/* empty */
;

axiomdef:	'(' AXIOM evtlvars	{ axiom=new PddlAxiom; axiom->SetVars($3); }
		CONTEXT goaldescr	{ axiom->SetContext($6); }
		IMPLIES literal		{ axiom->SetImplicand($9); }
		')'			{ $$=axiom; axiom=NULL; }
|		'(' AXIOM evtlvars	{ axiom=new PddlAxiom; axiom->SetVars($3); }
		IMPLIES literal		{ axiom->SetImplicand($6); }
		CONTEXT goaldescr	{ axiom->SetContext($9); }
		')'			{ $$=axiom; axiom=NULL; }
;

evtlvars:	VARS '(' typedvarset ')'	{ $$=$3; }
|		/* empty */			{ $$=new StringStringMap; }
;

/* ================================================================ */

addendumdef:    '(' DEFINE '(' ADDENDUM NAME ')'	{ UNIMPLEMENTED; }
			'(' DDOMAIN NAME ')'
			extradeflist ')'
;

extradeflist:	extradef extradeflist
|		/* empty */
;

extradef:	actiondef		{ }
|		axiomdef		{ }
|		safetydef		{ }
;

/* ================================================================ */

problemdef:     '(' DEFINE '(' PROBLEM NAME ')'	{ problem=new PddlProblem($5); }
		    probdeflist ')'
			{ yydatabase->RegisterProblem(problem); problem=NULL; }
;

probdeflist:	probdef probdeflist
|		/* empty */
;

probdef:	probrequiredef	{ problem->Requires($1); }
|		situation	{ problem->BasedOnSituation(*yydatabase->Situation(*$1)); delete $1; }
|		objectdecl	{ problem->SetObjects($1); }
|		init		{ problem->SetInit($1); }
|		goal		{ problem->SetGoal($1); }
|		lengthspec	{ }
|		probdnamedef	{ }
;

probdnamedef:	'(' DDOMAIN NAME ')'	{ problem->domainname=*$3;
					delete $3; };

probrequiredef:	'(' REQUIREMENTS reqsset ')'	{ $$=$3; };

situation:	'(' DSITUATION NAME ')'		{ $$=$3; };

objectdecl:	'(' CONSTANTS typednameset ')' { $$=$3; };

init:		'(' INIT literalset ')'		{ $$=$3; };

goal:		'(' GOAL goaldescr ')'		{ $$=$3; };

lengthspec:	'(' LENGTH seriallen parallellen ')' ;

seriallen:	'(' SERIAL NAME ')'	{ problem->serial=atoi($3->c_str()); delete $3; };

parallellen:	'(' PARALLEL NAME ')' { problem->parallel=atoi($3->c_str()); delete $3; };

/* ================================================================ */

situationdef:   '(' DEFINE '(' SITUATION NAME ')' { situation=new PddlSituation($5); }
		sitdeflist ')'
			{ yydatabase->RegisterSituation(situation); situation=NULL; }
;

sitdeflist:     sitdef	sitdeflist
|		/* empty */
;

sitdef:		sitdnamedef
|		sitobjectdecl
|		sitinitdef
;

sitdnamedef:	'(' DDOMAIN NAME ')'	{ situation->SetDomainname($3); } ;

sitobjectdecl:	'(' CONSTANTS typednameset ')' { situation->SetConstants($3); } ;

sitinitdef:	'(' INIT literalset ')' { situation->SetInit($3); } ;

/* ================================================================ */

goaldescr:      atom
|               equality
|		'(' AND goaldescrset ')'
			{ $$=new Term(MakeTerm(Term::And,*$3)); delete $3; }
|		'(' OR goaldescrset ')'
			{ $$=new Term(MakeTerm(Term::Or,*$3)); delete $3; }
|		'(' NOT goaldescr ')'
			{ $$=new Term(MakeTerm(Term::Not,*$3)); delete $3; }
|		'(' IMPLY goaldescr goaldescr ')'
			{ $$=new Term(MakeTerm(Term::Imply,*$3,*$4));
			  delete $3; delete $4; }
|		'(' EXISTS '(' typedvarset ')' goaldescr ')'
			{ $$=new Term(MakeTerm(Term::Exists, *$4,*$6));
			  delete $4; delete $6; }
|		'(' FORALL '(' typedvarset ')' goaldescr ')'
			{ $$=new Term(MakeTerm(Term::Forall, *$4,*$6));
			  delete $4; delete $6; }
|		'(' ISGOAL goaldescr ')'
			{ $$=new Term(MakeTerm(Term::Special,"IsGoal",*$3)); delete $3; }
|		'(' ISINITIAL goaldescr ')'
			{ $$=new Term(MakeTerm(Term::Special,"IsInitial",*$3)); delete $3; }
|		'(' LIMITED NAME goaldescr ')'
			{ $$=new Term(MakeTerm(Term::Special,string("Limit")+*$3,*$4)); delete $3; delete $4; }
;

goaldescrset:	goaldescr goaldescrset	{ $$=$2; $$->insert(*$1); delete $1; }
|		/* empty */		{ $$=new TermSet; }
;

effect:		atom
|		'(' AND effectset ')'
			{ $$=new Term(MakeTerm(Term::And,*$3)); delete $3; }
|		'(' NOT effect ')'
			{ $$=new Term(MakeTerm(Term::Not,*$3)); delete $3; }
|		'(' FORALL '(' typedvarset ')' effect ')'
			{ $$=new Term(MakeTerm(Term::Forall, *$4,*$6));
			  delete $4; delete $6; }
|		'(' WHEN goaldescr effect ')'
			{ $$=new Term(MakeTerm(Term::When,*$3,*$4));
			  delete $3; delete $4; }
;

effectset:	effect effectset
				{ $$=$2; $$->insert(*$1); delete $1; }
|		/* empty */	{ $$=new TermSet; }
;

/* ================================================================ */

reqsset:	REQUIREMENT reqsset     { $2->insert(*$1); $$=$2; delete $1; }
|		/*empty*/		{ $$=new StringSet; }
;

atom:           '(' NAME varornamelist ')'
			{ $$=new Term(MakeTerm(*$2,*$3)); delete $2; delete $3; }
;

equality:       '(' '=' VARNAME NAME ')'
			{ $$=new Term(MakeTerm(Term::Equal,*$3,*$4));
			  delete $3; delete $4; }
|		'(' '=' NAME VARNAME ')'
			{ $$=new Term(MakeTerm(Term::Equal,*$4,*$3));
			  delete $3; delete $4; }
|		'(' '=' VARNAME VARNAME ')'
			{ $$=new Term(MakeTerm(Term::Equal,*$3,*$4));
			  delete $3; delete $4; }
;

literalset:	literal literalset	{ $$=$2; $$->insert(*$1); delete $1; }
|		/* empty */		{ $$=new TermSet; }
;

literal:	atom			{ $$=$1; }
|		'(' NOT atom ')'	{ $$=new Term(MakeTerm(Term::Not,*$3));
					  delete $3; }
;

atomtypemap:	'(' NAME typedvarlist ')' atomtypemap
			{ $$=$5; (*$$)[*$2]=PddlMakeAtomtype($3); delete $2; }
|		/* empty */		{ $$=new StringAtomtypeMap; }
;

typednameset:	namelist		{ $$=PddlVarSetAddType($1,"object"); }
|		namelist '-' type typednameset { $$=PddlVarSetAddType($1,$3,$4); }
;

namelist:	NAME namelist		{ $$=$2; $2->push_front(*$1); delete $1; }
|		/* empty */		{ $$=new StringList; }
;

varornamelist:	NAME varornamelist      { $$=$2; $2->push_front(*$1); delete $1; }
|               VARNAME varornamelist   { $$=$2; $2->push_front(*$1); delete $1; }
|		/* empty */		{ $$=new StringList; }
;

typedvarset:	varlist			{ $$=PddlVarSetAddType($1,"object"); }
|		varlist '-' type typedvarset { $$=PddlVarSetAddType($1,$3,$4); }
;
	
typedvarlist:	varlist			{ $$=PddlVarListAddType($1,"object"); }
|		varlist '-' type typedvarlist { $$=PddlVarListAddType($1,$3,$4); }
;
	
varlist:	VARNAME varlist		{ $$=$2; $$->push_front(*$1); delete $1; }
|		/* empty */		{ $$=new StringList; }
;

type:           NAME                    { $$=$1; }
|		'(' EITHER namelist ')'	{ $$=PddlEitherID($3); }
|		'(' FLUENT type ')'	{ UNIMPLEMENTED; }
;

%%

static void shutoff_warnings()  /* does nothing but avoid compiler warningsa */
{
  problem=NULL;
  situation=NULL;
  if(0) shutoff_warnings();
}

extern const char *yyfilename;

int yyparsefile(const char *name)
{
  yydebug=0;
  yymylineno=1;
  yyfilename=name;
  yyin=fopen(name,"r");
  if(!yyin) {
    perror(name);
    return 2;
  }
  int res=yyparse();
  fclose(yyin);
  return res;
}

static void yyerror (char *s)  /* Called by yyparse on error */
{
  fprintf (stderr, "%s\n", s);
  printerrorlocation();
}

static StringStringPairList*
PddlVarListAddType(StringList* varlist, string type)
{
  return PddlVarListAddType(varlist, new string(type), new StringStringPairList);
}

static class makestringstringpair : public binary_function<const string,const string,StringStringPair >
{
public:
  result_type operator() (const string& s1, const string& s2) const
  {
    return result_type(s1,s2);
  }
} makestringstringpairobj;

static StringStringPairList*
PddlVarListAddType(StringList* varlist, string* type, StringStringPairList* rest)
{
  // :COMPILER: This does not work due to bug in my STL. Thu Dec  4 07:46:20 2003
  // transform(varlist->begin(), varlist->end(), inserter(*rest,rest->begin()),
  //           bind2nd(makestringstringpairobj, *type));
  // workaround:
  StringList::iterator it;
  for( it=varlist->begin(); varlist->end() != it; ++it)
    rest->push_front(makestringstringpairobj(*it,*type));
  delete varlist; delete type;
  return rest;
}

static StringStringMap*
PddlVarSetAddType(StringList* varlist, string type)
{
  return PddlVarSetAddType(varlist, new string(type), new StringStringMap);
}

static StringSet SplitEitherTypeAdd(string eithertype,StringSet res)
{
  string::size_type pre,post;
  pre=eithertype.find("(");
  post=eithertype.find_first_of(",)",pre+1,2);
  do {
    string type(eithertype,pre+1,post-pre-1);
    res.insert(type);
    pre=post;
    post=eithertype.find_first_of(",)",pre+1,2);
  } while (')'!=eithertype[pre]);
  return res;
}

static string PddlEitherAdd(string type1,string type2)
{
  StringSet     alltypes;
  if (type1.find("either(")!=0) alltypes.insert(type1);
  else alltypes=SplitEitherTypeAdd(type1,alltypes);
  if (type2.find("either(")!=0) alltypes.insert(type2);
  else alltypes=SplitEitherTypeAdd(type2,alltypes);
  string res=accumulatesep(alltypes.begin(),alltypes.end(),string("either("),string(","))+")";
  return res;
}

static StringStringMap*
PddlVarSetAddType(StringList* varlist, string* type, StringStringMap* rest)
{
  StringList::const_iterator it=varlist->begin(), end=varlist->end();
  for (;it!=end;++it) {
    if (rest->find(*it)==rest->end()) (*rest)[*it]=*type;
    else (*rest)[*it]=PddlEitherAdd((*rest)[*it],*type);
  }
  delete varlist; delete type;
  return rest;
}

static Atomtype PddlMakeAtomtype(StringStringPairList* typedvarlist)
{
  Atomtype res;
  for(StringStringPairList::iterator it=typedvarlist->begin();
      typedvarlist->end()!=it; ++it)
    res.push_back(it->second);
  delete typedvarlist;
  return res;
}

static string*  PddlEitherID(StringList* typelist)
{
  string* res=new string("either(");
  StringList::iterator it=typelist->begin();
  assert(typelist->end()!=it);
  *res+=*it;
  ++it;
  while (typelist->end()!=it) {
    *res+=",";
    *res+=*it;
    ++it;
  }
  *res+=")";
  delete typelist;
  return res;
}

  
