/* objcal.h	-- Hans-Peter Stoerr Tue Dec 22 1998
 * $Id: objcal.h,v 2.26 2003/12/04 08:20:07 stoerr Exp $
 */
/*  Objectoriented interface to cal-2.1 Bdd library.
 */

#ifndef _OBJCAL_H
#define _OBJCAL_H

#include <assert.h>
#include <string>

#include <map>
#include <vector>
#include <functional>

using namespace std;

// ================================================================
// Some extracts from cal.h: (You don't have to worry about this.)

typedef struct Cal_BddManagerStruct *Cal_BddManager;
typedef struct CalBddNodeStruct *Cal_Bdd;
typedef void *Cal_Pointer_t;

extern "C" {
extern void Cal_BddFree(Cal_BddManager BddManager, Cal_Bdd userBdd);
extern void Cal_BddUnFree(Cal_BddManager BddManager, Cal_Bdd userBdd);
}

#define CAL_REORDER_NONE 0
#define CAL_REORDER_SIFT 1
#define CAL_REORDER_WINDOW 2
#define CAL_REORDER_METHOD_BF 0
#define CAL_REORDER_METHOD_DF 1

// ================================================================
class BddMgr;
class BddAssoc;
class BddVector;

class Bdd
// this represents the BDDs.
{
  friend class BddMgr;
  friend class BddAssoc;
public:
  Bdd();                        // unassigned, invalid state.
  Bdd(const Bdd&);
  Bdd(const string);
  ~Bdd();

  // operators
  Bdd operator+(const Bdd) const;
  Bdd operator*(const Bdd) const;
  Bdd operator^(const Bdd) const;
  Bdd operator-() const;
  bool operator==(const Bdd) const;
  bool operator!=(const Bdd) const;

  Bdd& operator=(const Bdd&);
  Bdd& operator+=(const Bdd& b) { *this=*this+b; return *this; }
  Bdd& operator*=(const Bdd& b) { *this=*this*b; return *this; }
  Bdd& operator^=(const Bdd& b) { *this=*this^b; return *this; }

  Bdd   Implies(const Bdd b) const { return -(*this)+b; } // Implication
  Bdd   Equiv(const Bdd) const;   // Xnor
  friend Bdd Between(const Bdd,const Bdd);

  friend Bdd ITE(const Bdd,const Bdd,const Bdd);

  friend Bdd MultiwayOr(const BddVector &);
  friend Bdd MultiwayAnd(const BddVector &);
  friend Bdd MultiwayXor(const BddVector &);

  friend long Size(BddVector &); // returns joint size of Bdds.
  
  // tests
  bool operator<=(const Bdd) const;
  bool IsValid() const;
  bool IsVariable() const;
  bool IsCube() const;
  bool IsConst() const;
  bool IsOne() const;
  bool IsZero() const;
  bool Intersects(Bdd) const;   // returns true if (*this)*p != mgr.Zero()

  // Quantification
  Bdd Exists(const BddAssoc&) const;
  Bdd Exists(const Bdd) const;
  Bdd Exists(const BddVector&) const;
  Bdd ForAll(const BddAssoc&) const;
  Bdd ForAll(const Bdd) const;
  Bdd ForAll(const BddVector&) const;
  Bdd RelProd(const BddAssoc&, const Bdd) const; // Relation Product: And and Ex-Quant.
  Bdd RelProd(const Bdd exQuant, const Bdd) const;

  // Composition
  Bdd Compose(const Bdd var, const Bdd func) const; // replaces var by func
  Bdd SwapVars(const Bdd var1, const Bdd var2) const;
  Bdd Substitute(const BddAssoc&) const; // simultaneous composition
  
  // Constraints etc.
  Bdd Constrain(const Bdd c) const; // generalized cofactor wrt. c
  Bdd Reduce(const Bdd c) const; // returns poss. smaller Bdd which agrees on c

  // Set of Support etc.
  bool DependsOn(const Bdd var) const; // does *this depend on var?
  BddVector Support() const;       // Set of Support.
  Bdd Satisfy() const;          // returns some cube contained in *this. Deterministic.
  Bdd SatisfySupport(const BddAssoc&) const; // returns some cube contained in *this, with a node for
                                // each var. in assoc. Deterministic.
  Bdd RandomPath() const;       // returns random cube contained in *this. Not deterministic.
  Bdd SuperCube() const;        // returns smallest cube larger than *this, but with (at most) same vars.
  
  // File In/Out
  void Dump(const string name) const; // dump to File name
  static Bdd UnDump(const string name, int *error=NULL);
  
  // misc.
  string VarName() const;
  string Describe(int maxlen=10000) const;
  friend ostream& operator << (ostream& s, Bdd b);
  void Print(const char *name=NULL) const; // to file name name or stdout.
  void Profile(const char *name=NULL) const;
  void PrintInfo(const char *name=NULL) const; // prints nodecount, supportsize and totalsize
  long Size() const;

  Bdd IfPart() const;
  Bdd ThenPart() const;
  Bdd ElsePart() const;
  int IfIndex() const;          // Index of top variable
  int IfId() const;             // Id of top variable

  // int Hash() const;             // probably not stable under reordering
  int TotCmp(const Bdd) const;  // probably not stable under reordering

  // this is needed for BddBddMap.
  struct BddCompare : public binary_function<Bdd, Bdd, bool> {
    bool operator()(const Bdd& x, const Bdd& y) const { return x.TotCmp(y)<0; }
  };

  struct BddIndexCompare : public binary_function<Bdd, Bdd, bool> {
    bool operator()(const Bdd& x, const Bdd& y) const { return x.IfIndex()<y.IfIndex(); }
  }; // suitable for ordering variables only. Unstable under reordering!`
  struct BddIdCompare : public binary_function<Bdd, Bdd, bool> {
    bool operator()(const Bdd& x, const Bdd& y) const { return x.IfId()<y.IfId(); }
  }; // suitable for ordering variables only. Unstable under reordering!

protected:
  Bdd(Cal_Bdd);                 // does not unFree Bdd!
  typedef enum {unfree} unfreeT;
  Bdd(Cal_Bdd,unfreeT);         // does unFree Bdd
  operator Cal_Bdd() const;

  static bool printCubes(Bdd,string*,const string,bool,int maxlen=10000);
  static bool describe(Bdd, string *, const string, bool,int maxlen=10000);
  
  static map<Bdd,int,BddCompare>  sSingleVarAssocIds; // Assocs for each single variable
  int   MakeSingleVarAssocId(Bdd b) const;
  
  class BddRep
  // Wrapper around Cal_Bdd for internal use.
  // this solves the problem of automatically freeing and unfreeing Cal_Bdds.
  // Calls to functions look like this:
  // BddRep a=Cal_BddAnd(mgr,b,c)
  {
  public:
    BddRep(); // unassigned, invalid state.
    BddRep(const BddRep& b);
    BddRep& operator=(const BddRep& b);
    ~BddRep();

    // BddRep(const Cal_Bdd Bdd);  // this assumes Bdd is already unfreed.
    BddRep& operator=(const Cal_Bdd Bdd); // this assumes Bdd is already unfreed.
    operator Cal_Bdd() const;

    bool IsValid() const;
  private:
    Cal_Bdd               theBdd;
  };

  BddRep rep;

  void  p();                    // this exists only so I can call it from gdb.

};

// ________________________________________________________________
// non--member functions

// Bdd MultiwayAnd(const BddVector& set); // returns And of all Bdds of Set
// Bdd MultiwayOr(const BddVector& set); // returns Or of all Bdds of Set
// Bdd MultiwayXor(const BddVector& set); // returns Xor of all Bdds of Set

Bdd ITE(const Bdd ifB,const Bdd thenB , const Bdd elseB); // if then else operation
Bdd Between(const Bdd thisB, const Bdd thatB); // minimal Bdd between this and that.

long Size(BddVector&);     // returns joint size of Bdds.

// ================================================================
// general type definitions

typedef map<string, unsigned, less<string> > StringunsignedMap;
typedef map<unsigned, string, less<unsigned> > unsignedStringMap;
class BddVector : public vector<Bdd> {};
typedef map<Bdd, Bdd, Bdd::BddCompare> BddBddMap;

// ================================================================

class BddAssoc
{
public:
  BddAssoc();
  BddAssoc(const BddVector&);
  BddAssoc(const BddBddMap&);
  BddAssoc(const Bdd);          // only single Bdd in Assoc.
  ~BddAssoc();
  BddAssoc &operator=(const BddVector&);
  BddAssoc &operator=(const BddBddMap&);

  int   Id() const;
  bool  IsMapping() const;      // true if assoc is not an array but a mapping
  bool  MapsToVariables() const; // true if assoc maps only to variables

  void  SetToPrimedSet(int np); // all np-primed variables
  void  SetToPrimedSubst(int np1, int np2); // substitution of all np1-primed vars to all np2-primed vars.
  void  SetToSwitchSubst(int np1, int np2); // exchange of all np1-primed vars and all np2-primed vars.
private:
  int   assocId;
  bool  onlyVars;               // true if assoc contains only variables
  bool  isMap;                  // true if assoc is mapping, false if assoc is array.
  void  assign(const BddVector&);
  void  assign(const BddBddMap&);
  void  unassign();

  BddAssoc(const BddAssoc&) {abort();}; // not allowed.
  BddAssoc& operator=(const BddAssoc&) {abort();}; // not allowed.
};

// ================================================================

class BddMgr
{
  friend class Bdd;
  friend class Bdd::BddRep;
  friend class BddAssoc;
  
  friend ostream& operator << (ostream& s, Bdd b);
  friend Bdd Between(const Bdd,const Bdd);
  friend Bdd ITE(const Bdd,const Bdd,const Bdd);
  friend Bdd MultiwayOr(const BddVector &);
  friend Bdd MultiwayAnd(const BddVector &);
  friend Bdd MultiwayXor(const BddVector &);
  friend long Size(BddVector &); // returns joint size of Bdds.  
public:
  BddMgr();
  ~BddMgr();

  bool  Operative() { return NULL!=theManager; } // returns true if Manager works.
  
  Bdd   Var(const string);
  Bdd   One() const; Bdd True() const { return One(); }
  Bdd   Zero() const; Bdd False() const { return Zero(); }
  Bdd   Random(double satisFrac=0.5, BddVector support=BddVector()); // generate random Bdd
  Bdd   Random(BddVector support) { return Random(0.5,support); }
  // generate random Bdd from xor of ncubes random cubes
  // support defaults to mgr.NPrimedVars(0), ncubes to support.size()+10
  Bdd   XorRandom(BddVector support=BddVector(), int ncubes=0);
  Bdd   XorRandom(int ncubes) { return XorRandom(BddVector(),ncubes); }

  // misc functions
  long  PeakNumNodes();         // maximal number of nodes that has been allocated.
  long  GetNumNodes();          // number of nodes that is allocated.
  long  SetNodeLimit(long newlimit); // sets max. number of bdd nodes, returns old one.
  void  EnableGC(bool dogc);    // enables / disables garbage collection
  void  DoGC();                 // does a garbage collection
  bool  WasOverflow();          // returns true if there has been an overflow; overflow bit is cleared.
  void  SetOverflowFlag();        // forces us to behave like there was an overflow.
  
  void  Reorder(int method);    // invokes reordering
  void  SetDynReordering(int method); // allows dynamic reordering.
                                     
  // some BddAssociations etc.
  BddVector NPrimedVars(int N);
  const BddAssoc&       Unprimed(); // set of unprimed vars
  const BddAssoc&       Primed(); // set of primed vars
  const BddAssoc&       UnprimedToPrimed(); // substitute unprimed to primed vars
  const BddAssoc&       PrimedToUnprimed(); // substitute primed vars to unprimed

protected:
  operator Cal_BddManager() const;
  string varName(Cal_Bdd);

private:
  Cal_BddManager        theManager;

  StringunsignedMap  varNameIdMap; // variable Id from variable name.
  unsignedStringMap  varIdNameMap; // variable name from Id.

  bool  mAssocsValid;
  BddAssoc noprimed, primed;
  BddAssoc noprimetoprime, primetonoprime;
  bool  mForcedOverflowFlag;    // if set we will behave like there was an overflow
  
  void  updateAssocs();
  void insertVarInMaps(Bdd,string);
  static char* printNameVars(Cal_BddManager mgr, Cal_Bdd varpt, Cal_Pointer_t mgrobj);
};

// THE Bdd-manager. Do not declare more! This is hardcoded into Bdd!
extern BddMgr mgr;              

#endif
